import time
import random

URL_TEMPLATE = 'https://raw.githubusercontent.com/microprediction/precisedata/main/returns/fathom_data_N.csv'

n_data = 450

from timemachines.inclusion.pandasinclusion import using_pandas

if using_pandas:
    import pandas as pd

    def random_long_data(n_obs):
        """ Returns random long time series with reasonable amount of signal """
        assert n_obs<35000

        start_time = time.time()
        t = [ start_time+24*60*60*60*i for i in range(n_obs) ]

        got = False
        while not got:
            the_choice = random.choice(list(range(n_data)))
            the_url = URL_TEMPLATE.replace('N', str(the_choice))
            try:
                df = pd.read_csv(the_url)
                del df['Unnamed: 0']
                got = len(df.index) > n_obs+10
            except:
                got = False
        col = random.choice(list(df.columns))
        vals = list(df[col].values)
        k = random.choice(list(range(0,len(df.index)-n_obs-5)))
        y = vals[k:k+n_obs]
        return y,t


if __name__=='__main__':
    y,t = random_long_data(n_obs=1000)
    print(y[:10])


from timemachines.skaters.kts.ktsinclusion import using_kats

if using_kats:
    import numpy as np
    import pandas as pd
    from typing import List
    from kats.consts import TimeSeriesData
    from kats.models.prophet import ProphetModel, ProphetParams
    from kats.models.holtwinters import HoltWintersModel, HoltWintersParams
    from kats.models.quadratic_model import QuadraticModel, QuadraticModelParams


    def kats_Prophet_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Kats' Prophet forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'}).reset_index()
        df.columns = ["time", "value"]

        train_data = TimeSeriesData(df.head(df.shape[0]))

        params = ProphetParams(seasonality_mode='multiplicative') # additive mode gives worse results
        m = ProphetModel(train_data, params)
        m.fit()
        fcst = m.predict(steps=k, freq="H")

        x = list(fcst.fcst.values)
        x_std = [1]*k

        return x, x_std

    if True:
        def kats_HoltWinters_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
            """
                Calls Kats' Holt-Winters forecasting model, but ignores t if supplied.
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [ yt for yt in y]
            else:
                y0s = [ yt[0] for yt in y ]

            idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
            df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'}).reset_index()
            df.columns = ["time", "value"]

            train_data = TimeSeriesData(df.head(df.shape[0]))

            params = HoltWintersParams(
                        trend="add"
                        #damped=False,
                        #seasonal="mul",
                        #seasonal_periods=12,
                    )

            m = HoltWintersModel(
                data=train_data, 
                params=params
                )

            m.fit()
            fcst = m.predict(steps=k, alpha = 0.5)

            x = list(fcst.fcst.values)
            x_std = [1]*k

            return x, x_std

    if True:
        def kats_quadratic_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
            """
                Calls Kats' quadratic regressions forecasting model, but ignores t if supplied.
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [ yt for yt in y]
            else:
                y0s = [ yt[0] for yt in y ]

            idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
            df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'}).reset_index()
            df.columns = ["time", "value"]

            train_data = TimeSeriesData(df.head(df.shape[0]))

            params = QuadraticModelParams()

            m = QuadraticModel(train_data, params)
            m.fit()
            fcst = m.predict(steps = k)

            x = list(fcst.fcst.values)
            x_std = [1]*k

            return x, x_std


if __name__=='__main__':
    assert using_kats, 'pip install kats'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = kats_Prophet_iskater(y=y, k=5)
    print(x)
    print(time.time()-st)


from timemachines.skatertools.utilities.conventions import Y_TYPE, from_space, A_TYPE

from timemachines.inclusion.sklearninclusion import using_sklearn
from timemachines.inclusion.humpdayinclusion import using_humpday
if using_sklearn and using_humpday:

    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error_with_sporadic_fit
    import traceback
    import time


    # Find the optimal hyper-parameter r for a skater

    # TODO:Tidy this up



    def infer_r_dim(skater_name):
        """ A convention for skater names """
        return 3 if '_r3' in skater_name else 2 if '_r2' in skater_name else 1 if '_r1'in skater_name else None


    def optimal_r(f, y:[Y_TYPE],
                  k:int, a:[A_TYPE], t=None, e=None,
                  evaluator=None, optimizer=None,
                  n_trials=None, n_dim=None,  # Which dimension to search in
                  n_burn:int=None,
                  test_objective_first=True)->(float,float,dict):
        """
              Returns:  best_r: float
                        best_val: float
                        info: {best_rn: interpretation in n_dim space of r,
                               feval_count: number of actual objective evaluations,
                              }

        """
        if not using_humpday:
            raise Exception('You must manually install humpday, with pip install humpday, in order to use this')
        if evaluator is None:
            evaluator = evaluate_mean_squared_error_with_sporadic_fit

        if n_dim is None:
            n_dim = infer_r_dim(f.__name__)
            assert n_dim is not None, "Hyper-parameter dimension was not supplied and could not be inferred from "+str(f.__name__)

        assert n_burn is not None, "Must supply n_burn"

        def objective(u:[float]):
            r = from_space(u)
            return evaluator(f=f, y=y, k=k, a=a, t=t, r=r)

        if test_objective_first:
            start_time = time.time()
            a_test = objective(u=[0.5]*n_dim)  # Fail fast with easier trace
            elapsed = time.time()-start_time
            print('One evaluation took '+ str(round(elapsed/60,1))+' minutes.')
            print('Come back in '+str(n_trials*elapsed/60)+' minutes.')
        best_val, best_rn, feval_count = optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True)
        best_r = from_space(best_rn)
        info = {'best_rn':best_rn,'feval_count':feval_count}
        return best_r, best_val, info




    from timemachines.skaters.proph.prophetinclusion import using_prophet
    # Just an example ... probably belongs elsewhere
    if using_prophet:
        def optimize_proph_against_hospital(k=10, n_extra=100):
            """ Illustrative example

                   n_extra is roughly the number of times the prophet model will be called

            """
            assert using_humpday,'pip install humpday'
            from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error_with_sporadic_fit
            from timemachines.skatertools.data.real import hospital_with_exog
            from timemachines.skaters.proph.prophparams import PROPHET_META
            from humpday.optimizers.optunacube import optuna_tpe_cube
            evaluator = evaluate_mean_squared_error_with_sporadic_fit
            n_burn = PROPHET_META['n_warm'] + k + 1
            y, a = hospital_with_exog(k=k, n=n_burn+n_extra, offset=True)

            from timemachines.skaters.proph.prophskaterssingular import fbprophet_univariate, fbprophet_known, fbprophet_recursive
            from timemachines.skaters.simple.movingaverage import EMA_SKATERS
            skaters = [fbprophet_univariate, fbprophet_known, fbprophet_recursive, fbprophet_recursive] + EMA_SKATERS

            for f in skaters:
                mn_err = evaluator(f=f, y=y, k=k, a=a, n_burn=n_burn)
                print(f.__name__ + ' : '+ str(mn_err))

            # Finally, try to optimize prophet with hyper-parameters
            from timemachines.skaters.proph.prophskaterssingular import fbprophet_exogenous_r2
            f = fbprophet_exogenous_r2
            n_dim = 2
            best_r, best_val, info = optimal_r(f=f, y=y, k=k, a=a, evaluator=evaluator,
                                         optimizer=optuna_tpe_cube, n_burn=n_burn,
                                         n_trials=50, n_dim=n_dim,
                                         test_objective_first=True)
            from pprint import pprint
            pprint(info)
            print('Best r is ' + str(best_r))
            print('Best value is '+str(best_val))
            from timemachines.skaters.proph.prophparams import prophet_param
            from timemachines.skatertools.utilities.conventions import to_space
            rn = to_space(best_r)
            pprint({'changepoint prior':prophet_param(param_name='changepoint_prior_scale',u=rn[0]),
                   'seasonality prior':prophet_param(param_name='seasonality_prior_scale',u=rn[0])})




    def select_working_combinations():
        """ Tedious examination of which optimizers work against different evaluators """
        from timemachines.skatertools.data.synthetic import brownian_with_exogenous
        from timemachines.skatertools.evaluation.evaluators import EVALUATORS
        from humpday.optimizers.allotpimizers import OPTIMIZERS
        broken = list()
        k = 1
        from timemachines.skaters.simple.movingaverage import empirical_ema_r1
        f = empirical_ema_r1

        for optimizer in OPTIMIZERS:
            print(' ')
            print(optimizer.__name__)
            for evaluator in EVALUATORS:
                try:
                    print(evaluator.__name__, optimal_r(f=f, y=brownian_with_exogenous(n=40), k=k, a=None,
                                                        n_trials=5, n_dim=3, n_burn=20, optimizer=optimizer,
                                                        evaluator=evaluator, with_count=True))
                except Exception as e:
                    traceback.print_tb()
                    broken.append((optimizer.__name__, evaluator.__name__))
        print(' ')
        print('Broken : ')
        print(broken)


if __name__=='__main__':
    assert using_humpday
    assert using_sklearn
    assert using_prophet
    optimize_proph_against_hospital()





from timemachines.skaters.mrln.merlionarimaskater import MERLION_ARIMA_SKATERS
from timemachines.skaters.mrln.merlionmsesskater import MERLION_MSES_SKATERS
from timemachines.skaters.mrln.merlionprophetskater import MERLION_PROPHET_SKATERS

MERLION_SKATERS = MERLION_ARIMA_SKATERS + MERLION_MSES_SKATERS + MERLION_PROPHET_SKATERS




from timemachines.skaters.kts.ktsprophetskater import KATS_PROPHET_SKATERS
from timemachines.skaters.kts.ktsholtwintersskater import KATS_HOLT_WINTERS_SKATERS
from timemachines.skaters.kts.ktsquadraticskater import KATS_QUADRATIC_SKATERS

KATS_SKATERS = KATS_PROPHET_SKATERS + KATS_HOLT_WINTERS_SKATERS + KATS_QUADRATIC_SKATERS



# Compute performance rating for a single algorithm

# Not done

def performance_rating(skater_names:[str]):
    # Play in a small pool

    benchmark_skater_names = ['merlion_arima','tsa_p2_d0_q0','sluggish_moving_average','thinking_slow_and_fast',
                              'aggressive_ema_ensemble','tsa_balanced_theta_ensemble','thinking_fast_and_fast','empirical_last_value',
                              'rapidly_moving_average']
    from timemachines.skaters.allskaters import skater_from_name
    skater_population = [ skater_from_name(nm) for nm in benchmark_skater_names+skater_names ]
    from timemachines.skatertools.comparison.skaterelo import skater_elo_multi_update





from timemachines.skaters.pycrt.pycaretinclusion import using_pycaret


if using_pycaret:
    from timemachines.skaters.pycrt.pycaretwrapper import pycrt_iskater
    from timemachines.skatertools.batch.batchskater import batch_skater_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE


    def pycrt_median_3(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                       emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=pycrt_iskater,
                                    iskater_kwargs={'n_select': 3,'blend_method':'median'},
                                    min_e=0)


    def pycrt_median_3_full(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                       emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=pycrt_iskater,
                                    iskater_kwargs={'n_select': 3, 'blend_method': 'median','turbo':False},
                                    min_e=0)


    def pycrt_median_8(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=pycrt_iskater,
                                    iskater_kwargs={'n_select':8,'blend_method':'median'},
                                    min_e=0)

    def pycrt_mean_3(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=pycrt_iskater,
                                    iskater_kwargs={'n_select':3,'blend_method':'mean'},
                                    min_e=0)

    def pycrt_mean_8(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=pycrt_iskater,
                                    iskater_kwargs={'n_select':8,'blend_method':'mean'},
                                    min_e=0)

    PYCRT_MEDIAN_SKATERS = [pycrt_median_3, pycrt_median_8, pycrt_mean_3, pycrt_mean_8, pycrt_median_3_full]

else:
    PYCRT_MEDIAN_SKATERS = []

if __name__=='__main__':
    from timemachines.skatertools.evaluation.evaluators import hospital_mean_square_error_with_sporadic_fit
    err1 = hospital_mean_square_error_with_sporadic_fit(f=pycrt_median_3, k=3, n=110)



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any, List
from timemachines.skatertools.ensembling.precisionweightedskater import precision_weighted_skater
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import noneupdatelist

# Some r values to use when ensembling
R_PRECISION = 0.5    # (i.e. precision weighting)
R_BALANCED = 0.25
R_AGGRESSIVE = 0.90


def ensemble_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                     fs: List = None, rs: List = None, g=None, r=None, include_std=True, trust=True,
                     empirical_std=False) -> ([float], Any, Any):
    """ Meta-model skater

          fs  - list of skaters
          rs  - list of hyper-params, if any
          g   - exogenous skater (the meta model)
          r   - hyper-param for g, if any
          include_std   - If True, will interleave x_std in the exogenous variables sent to g
          trust         - If True will use the models' standard deviation, otherwise empirical
          empirical_std - If True will use empirical std, otherwise leans on models' estimates

        It is possible to change the list of skaters supplied every now and then but be aware this will
        throw off the tracking of empirical errors in the parade a little.

    """
    if not s.get('s_fs'):
        s = {'s_fs': dict([(f.__name__,{}) for f in fs]),  # State for skaters
             's_g': {},                 # State for meta-model
             'n_obs':0,
             's_fs_p':dict([(f.__name__,{}) for f in fs]), # Holds parades for skaters
             's_p':{},                  # Self parade
             's_w':''                   # Holds current winning model
             }

    if y is None:
        return None, None, s
    else:
        # New f introduced?
        for f in fs:
            if not f.__name__ in s['s_fs']:
                s['s_fs'][f.__name__]={}
                s['s_fs_p'][f.__name__]={}

        # Apply models, keeping only the point estimate
        # This constructs a flat list of child predictions and their standard deviations,
        # but for now only uses the k-step ahead predictions to judge, not k-1, k-2,...
        xjs = list()
        rs = rs or [None for _ in fs]
        y0 = wrap(y)[0]
        s['s_w'] = fs[0].__name__
        lowest_std = 1e14
        for j, (f, r) in enumerate(zip(fs, rs)):
            if r is not None:
                xj, xj_std, s['s_fs'][f.__name__] = f(y=y, s=s['s_fs'][f.__name__], k=k, a=a, t=t, e=e, r=r)
            else:
                xj, xj_std, s['s_fs'][f.__name__] = f(y=y, s=s['s_fs'][f.__name__], k=k, a=a, t=t, e=e)
            xjs.append(xj[-1])  # <--- Only use the k-step forward prediction and toss out the rest (a bit lazy)

            if not trust:
                _, emp_std, s['s_fs_p'][f.__name__] = parade(p=s['s_fs_p'][f.__name__], x=xj, y=y0 )
                xj_std = noneupdatelist(xj_std, emp_std)

            if xj_std[-1] < lowest_std:
                s['s_w'] = f.__name__
                lowest_std = xj_std[-1]
            if include_std:
                xjs.append(xj_std[-1])

        # Next we'll pass the child predictions to a combining meta-model
        s['n_obs']+=1
        if s['n_obs']<10:
            return [wrap(y)[0]]*k, [wrap(y)[0]]*k, s
        else:
            y_extend = [wrap(y)[0]] + xjs
            if r is None:
                x, x_std, s['s_g'] = g(y=y_extend, s=s['s_g'], k=k, a=a, t=t, e=e)
            else:
                x, x_std, s['s_g'] = g(y=y_extend, s=s['s_g'], k=k, a=a, t=t, e=e, r=r)
            if empirical_std:
                _, self_emp_std, s['s_p'] = parade(p=s['s_p'], x=x, y=y0)
                x_std = noneupdatelist(x_std, self_emp_std)
            return x, x_std, s


def precision_weighted_ensemble_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                     fs: List = None, rs: List = None, r=None, trust=False, empirical_std=True) -> ([float], Any, Any):
    """
             r -   determines the exponent
    """
    g = precision_weighted_skater
    return ensemble_factory(y=y,s=s,k=k,a=a,t=t,e=e,fs=fs,rs=rs,g=g,r=r, trust=trust, empirical_std=empirical_std)


def trusting_precision_weighted_ensemble_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None,
                     fs: List = None, rs: List = None, r=None, trust=True, empirical_std=True) -> ([float], Any, Any):
    """
             r -   determines the exponent
    """
    g = precision_weighted_skater
    return ensemble_factory(y=y,s=s,k=k,a=a,t=t,e=e,fs=fs,rs=rs,g=g,r=r, trust=trust, empirical_std=empirical_std)



from timemachines.skaters.mrln.merlioninclusion import using_merlion

if using_merlion:
    import numpy as np
    import pandas as pd
    from typing import List
    from merlion.utils.time_series import TimeSeries
    from merlion.models.forecast.prophet import Prophet, ProphetConfig
    from merlion.models.forecast.smoother import MSES, MSESConfig
    from merlion.models.forecast.arima import Arima, ArimaConfig

    def merlion_ARIMA_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Merlion's ARIMA forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)

        train_val_list = ([1] * len(y)) + ([0] * k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'})

        train_data = TimeSeries.from_pd(df.head(df.shape[0]-k))
        test_data = TimeSeries.from_pd(df.tail(k))

        model = Arima(ArimaConfig())
        model.train(train_data=train_data)
        test_pred, test_err = model.forecast(time_stamps=test_data.time_stamps)


        x = test_pred.to_pd()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std

    if True:
        # TODO: FIXME: This ain''t working
        def merlion_MSES_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
            """
                Calls Merlion's MSES forecasting model, but ignores t if supplied.
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [ yt for yt in y]
            else:
                y0s = [ yt[0] for yt in y ]

            y0s.extend([0]*k)

            train_val_list = ([1] * len(y)) + ([0] * k)
            idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
            df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'})

            train_data = TimeSeries.from_pd(df.head(df.shape[0]-k))
            test_data = TimeSeries.from_pd(df.tail(k))

            model = MSES(MSESConfig(k))
            model.train(train_data=train_data)
            test_pred, test_err = model.forecast(time_stamps=test_data.time_stamps)

            x = test_pred.to_pd()['y'].values.tolist()
            x_std = [1]*k

            return x, x_std

    if True:
        def merlion_Prophet_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
            """
                Calls Merlion's implemenetation of the Prophet forecasting model, but ignores t if supplied.
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [ yt for yt in y]
            else:
                y0s = [ yt[0] for yt in y ]

            y.extend([0]*k)

            train_val_list = ([1] * len(y)) + ([0] * k)
            idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
            df = pd.DataFrame(y0s, index=idx).rename(columns={0:'y'})

            train_data = TimeSeries.from_pd(df.head(df.shape[0]-k))
            test_data = TimeSeries.from_pd(df.tail(k))

            model = Prophet(ProphetConfig())
            model.train(train_data=train_data)
            test_pred, test_err = model.forecast(time_stamps=test_data.time_stamps)

            x = test_pred.to_pd()['y'].values.tolist()
            x_std = [1]*k

            return x, x_std


if __name__=='__main__':
    assert using_merlion, 'pip install salesforce-merlion'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = merlion_ARIMA_iskater(y=y, k=5)
    print(x)
    print(time.time()-st)


import time
import random

SKATER_RESIDUAL_URL = 'https://raw.githubusercontent.com/microprediction/precisedata/main/skaterresiduals/skater_residuals_0.csv'

n_data = 450

from timemachines.inclusion.pandasinclusion import using_pandas

if using_pandas:
    import pandas as pd

    def random_skater_residual_dataframe(n_obs:int):
        got = False
        while not got:
            the_choice = random.choice(list(range(n_data)))
            the_url = SKATER_RESIDUAL_URL.replace('N', str(the_choice))
            try:
                df = pd.read_csv(the_url)
                del df['Unnamed: 0']
                got = len(df.index) > n_obs + 10
            except:
                got = False
        return df

    def random_long_residual(n_obs):
        """ Returns random long time series of skater residuals """
        assert n_obs<35000

        start_time = time.time()
        t = [ start_time+24*60*60*60*i for i in range(n_obs) ]
        df = random_skater_residual_dataframe(n_obs)
        col = random.choice(list(df.columns))
        vals = list(df[col].values)
        k = random.choice(list(range(0,len(df.index)-n_obs-5)))
        y = vals[k:k+n_obs]
        return y,t

    def random_multivariate_residual(n_obs, as_dataframe=True, random_start=True):
        """ Canned skater residual data, potentially useful for studying ensembles """
        assert n_obs < 35000
        start_time = time.time()
        t = [start_time + 24 * 60 * 60 * 60 * i for i in range(n_obs)]
        df = random_skater_residual_dataframe(n_obs=n_obs)
        if random_start:
            k = random.choice(list(range(0, len(df.index) - n_obs - 5)))
            df = df[k:k+n_obs]
        return df[:n_obs] if as_dataframe else df[:n_obs].values


if __name__=='__main__':
    y,t = random_long_residual(n_obs=1000)
    print(y[:10])


try:
    from microprediction import MicroReader
    using_micro = True
except ImportError:
    using_micro = False

import random
import time

# Live data streams at www.microprediction.com

if using_micro:
    import numpy as np

    def random_stream_name_url_lagged(min_len=500, exclude_str=None, include_str=None,
                                      with_lagged=False,
                                      with_url=True,
                                      backoff_seconds=1):
        """ Randomly selected univariate series
                :returns:  name, url, lagged if with_lagged=True
        """
        mr = MicroReader()
        names = mr.get_stream_names()
        okay = False
        while not okay:
            name = random.choice(names)
            okay = True
            if exclude_str is not None and exclude_str in name:
                okay = False
            if include_str is not None and include_str not in name:
                okay = False
            if okay and min_len>0:
                lagged = mr.get_lagged(name=name, count=10000)
                n_obs = len(lagged)
                okay = n_obs >= min_len
                if not okay:
                    time.sleep(backoff_seconds)
        url = 'https://www.microprediction.org/stream_dashboard.html?stream=' + name.replace('.json', '')
        if with_lagged and with_url:
            return name, url, lagged
        elif with_url and (not with_lagged):
            return name, url
        elif (not with_url) and (with_lagged):
            return name, lagged
        elif (not with_url) and (not with_lagged):
            return name


    def random_regular_stream_name(min_len=0, with_url=False):
        """ Randomly selected univariate series with some data  """
        return random_stream_name_url_lagged(min_len=min_len, exclude_str='~',with_url=with_url, with_lagged=False)


    def random_chronological_values_and_times(min_len=500, exclude_str=None, include_str=None, backoff_seconds=1):
       _name, _url, lagged = random_stream_name_url_lagged(min_len=min_len, exclude_str=exclude_str, include_str=include_str,
                                                         with_lagged=True, backoff_seconds=backoff_seconds)
       # Chronological
       lagged_values = [l[1] for l in reversed(lagged)]
       lagged_times = [l[0] for l in reversed(lagged)]
       return lagged_values, lagged_times

    random_stream_data = random_chronological_values_and_times

    def stream_data(name:str,n_obs:int):
        """ values and times for a univariate stream """
        mr = MicroReader()
        lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name, count=n_obs)
        y, t = list(reversed(lagged_values)), list(reversed(lagged_times))
        return y, t


    def random_regular_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y,t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~')
        return y[:n_obs],t[:n_obs]

    def random_electricity_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='electricity')
        return y[:n_obs], t[:n_obs]

    def random_hospital_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='hospital')
        return y[:n_obs], t[:n_obs]

    def random_airport_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='airport')
        return y[:n_obs], t[:n_obs]

    def random_volume_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='volume')
        return y[:n_obs], t[:n_obs]

    def random_emoji_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='emoji')
        return y[:n_obs], t[:n_obs]


    def random_noaa_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='noaa')
        return y[:n_obs], t[:n_obs]


    def random_elo_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        data_source = random.choice([random_airport_data, random_electricity_data, random_noaa_data, random_emoji_data])
        return data_source(n_obs=n_obs)


    def random_electricity_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
            Chronological
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, exclude_str='~', include_str='electricity')
        return y[:n_obs], t[:n_obs]


    def random_residual_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        y, t = random_chronological_values_and_times(min_len=n_obs, include_str='~')
        return y[:n_obs], t[:n_obs]


    def random_rescaled_chronological_values_and_times(n_obs=900, include_str='electricity', exclude_str='~'):
        y, t = random_chronological_values_and_times(min_len=n_obs, include_str=include_str, exclude_str=exclude_str)
        scale = (max(y) -min(y) )+0.001
        return [ yt/scale for yt in y[:n_obs] ], t[:n_obs]


    def random_surrogate_data(f, k=1, n_real=50, n_samples=200, n_warm = 500, n_input=20, verbose=False,
                               include_str='electricity',exclude_str='~'):
        """ Create dataset of skater input (abbreviated to last n_input lags) and output
        :param f:   skater
        :param k:   step ahead
        :param n_real:   number of distinct real time series to use
        :param n_input:  number of lags to use in training
        :param n_warm:   number of data points to use to warm up skater before harvesting predictions
        :return: x_train, y_train, y_true
        """
        n_obs = n_samples + n_warm + 2* k + 10
        n_train = n_real * n_samples
        x_train = np.zeros(shape=(n_train, n_input, 1))
        y_train = np.zeros(shape=(n_train, 1))
        x_stds = [1.0]
        y_true = list()
        if verbose:
            print(np.shape(x_train))
        for ndx_real in range(n_real):
            # Get real data
            y_real, _ = random_rescaled_chronological_values_and_times(n_obs=n_obs, include_str=include_str, exclude_str=exclude_str)
            if verbose:
                print(y_real[:5])
                print((ndx_real, n_real))
                print(np.mean(x_stds))
            # Warm the model up
            y_real_warm = y_real[:n_warm]
            s = {}
            for y in y_real_warm:
                x, x_std, s = f(y, k=k, s=s)
            # Now create examples of k-step ahead forecasts
            y_real_harvest = y_real[n_warm:n_warm + n_samples]
            for j2, y in enumerate(y_real_harvest):
                x, x_std, s = f(y, k=k, s=s)
                xk = x[k - 1]
                j = ndx_real * n_samples + j2
                yj_true = y_real[n_warm + j2 + k]
                y_true.append(yj_true)
                y_train[j, 0] = xk
                x_stds.append(x_std[-k])
                reverse_input_data = y_real[n_warm + j2 - n_input+1:n_warm +j2+1]
                input_data = reverse_input_data
                #assert input_data[0]==y
                for l in range(n_input):
                    x_train[j, l, 0] = input_data[l]

        return x_train, y_train, y_true


if __name__=='__main__':
    if using_micro:
        name, url = random_regular_stream_name(min_len=900,with_url=True)
        print(url)


try:
   from statsmodels.tsa.arima.model import ARIMA
   using_statsmodels = True
except ImportError:
    try:
       from statsmodels.tsa.arima_model import ARIMA
       using_statsmodels = True
    except:
       using_statsmodels = False



# type-k-seconds


# my_skaters = top_rated_models(k=1, n=15, max_seconds=1, category='univariate')
U_1_1 = ['slow_precision_ema_ensemble', 'quick_precision_ema_ensemble', 'slow_aggressive_ema_ensemble', 'balanced_ema_ensemble', 'precision_ema_ensemble', 'quick_balanced_ema_ensemble', 'quick_aggressive_ema_ensemble', 'slow_balanced_ema_ensemble', 'sk_theta', 'aggressive_ema_ensemble', 'thinking_fast_and_fast', 'thinking_precision_ensemble', 'thinking_fast_and_slow', 'quickly_moving_average', 'slowly_moving_average']
#  my_skaters = top_rated_models(k=1, n=15, max_seconds=5,category='univariate')
U_1_5 = ['slow_precision_ema_ensemble', 'quick_precision_ema_ensemble', 'slow_aggressive_ema_ensemble', 'balanced_ema_ensemble', 'precision_ema_ensemble', 'tsa_aggressive_theta_ensemble', 'quick_balanced_ema_ensemble', 'tsa_balanced_theta_ensemble', 'quick_aggressive_ema_ensemble', 'slow_balanced_ema_ensemble', 'sk_theta', 'aggressive_ema_ensemble', 'tsa_precision_theta_ensemble', 'thinking_fast_and_fast', 'thinking_precision_ensemble']

U_3_1 = ['quick_aggressive_ema_ensemble', 'slow_precision_ema_ensemble', 'sluggish_moving_average', 'precision_ema_ensemble', 'aggressive_ema_ensemble', 'thinking_slow_and_fast', 'sk_theta', 'slow_aggressive_ema_ensemble', 'quickly_moving_average', 'balanced_ema_ensemble', 'thinking_slow_and_slow', 'quick_precision_ema_ensemble', 'slow_balanced_ema_ensemble', 'quick_balanced_ema_ensemble', 'slowly_moving_average']
U_8_1 = ['slow_precision_ema_ensemble', 'quick_precision_ema_ensemble', 'sluggish_moving_average', 'precision_ema_ensemble', 'thinking_slow_and_slow', 'slow_balanced_ema_ensemble', 'slow_aggressive_ema_ensemble', 'thinking_slow_and_fast', 'balanced_ema_ensemble', 'slowly_moving_average', 'aggressive_ema_ensemble', 'sk_theta', 'quickly_moving_average', 'quick_aggressive_ema_ensemble', 'thinking_fast_and_slow']
U_21_1 = ['thinking_slow_and_slow', 'sluggish_moving_average', 'slowly_moving_average', 'thinking_precision_ensemble', 'thinking_fast_and_slow', 'precision_ema_ensemble', 'thinking_slow_and_fast', 'quick_aggressive_ema_ensemble', 'slow_precision_ema_ensemble', 'pmd_univariate', 'quick_balanced_ema_ensemble', 'balanced_ema_ensemble', 'slow_balanced_ema_ensemble', 'quick_precision_ema_ensemble', 'aggressive_ema_ensemble']





#  my_skaters = top_rated_models(k=1, n=15, max_seconds=1,category='residual')
R_1_1 = ['quick_precision_ema_ensemble', 'thinking_fast_and_fast', 'precision_ema_ensemble', 'slow_precision_ema_ensemble', 'aggressive_ema_ensemble', 'balanced_ema_ensemble', 'slow_aggressive_ema_ensemble', 'slow_balanced_ema_ensemble', 'thinking_precision_ensemble', 'rapidly_moving_average', 'quick_aggressive_ema_ensemble', 'quickly_moving_average', 'quick_balanced_ema_ensemble', 'sk_theta', 'thinking_fast_and_slow']
# my_skaters = top_rated_models(k=1, n=15, max_seconds=5,category='residual')
R_1_5 = ['quick_precision_ema_ensemble', 'thinking_fast_and_fast', 'precision_ema_ensemble', 'tsa_precision_theta_ensemble', 'slow_precision_ema_ensemble', 'aggressive_ema_ensemble', 'balanced_ema_ensemble', 'slow_aggressive_ema_ensemble', 'tsa_balanced_theta_ensemble', 'slow_balanced_ema_ensemble', 'thinking_precision_ensemble', 'tsa_aggressive_theta_ensemble', 'rapidly_moving_average', 'quick_aggressive_ema_ensemble', 'quickly_moving_average']






from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
from timemachines.inclusion.statsmodelsinclusion import using_statsmodels

if using_statsmodels:

    from statsmodels.tsa.forecasting.theta import ThetaModel
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
    from typing import Any
    from timemachines.skaters.tsa.tsaparams import TSA_META
    from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
    from timemachines.skatertools.components.parade import parade
    from timemachines.skatertools.utilities.nonemath import nonecast
    import warnings
    from statsmodels.tools.sm_exceptions import ConvergenceWarning
    warnings.simplefilter('ignore', ConvergenceWarning)


    def tsa_theta_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None,
                          method=None, deseasonalize=True,
                          use_test=True, difference=False) -> ([float], Any, Any):
        """

        period : int, default None
            The period of the data that is used in the seasonality test and
            adjustment. If None then the period is determined from y's index,
            if available.
        deseasonalize : bool, default True
            A flag indicating whether the deseasonalize the data. If True and
            use_test is True, the data is only deseasonalized if the null of no
            seasonal component is rejected.
        use_test : bool, default True
            A flag indicating whether test the period-th autocorrelation. If this
            test rejects using a size of 10%, then decomposition is used. Set to
            False to skip the test.
        method : {"auto", "additive", "multiplicative"}, default "auto"
            The model used for the seasonal decomposition. "auto" uses a
            multiplicative if y is non-negative and all estimated seasonal
            components are positive. If either of these conditions is False,
            then it uses an additive decomposition.
        difference : bool, default False
            A flag indicating to difference the data before offlinetesting for
            seasonality."""

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'y': list(),
                 'a': list(),
                 'k': k,
                 'p':{}}
        else:
            # Assert immutability of k, dimensions
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if len(s['y']) > max(2 * k + 5, TSA_META['n_warm']) and (e is not None and e>0):
                y0s = [ y_[0] for y_ in s['y']]
                import numpy as np
                model = ThetaModel(np.array(y0s), method=method, deseasonalize=deseasonalize,
                          use_test=use_test, difference=difference)
                try:
                    x = list( model.fit().forecast(steps=k) )
                except Exception as e:
                    print(e)
                    x = [wrap(y)[0]]*k
            else:
                x = [y[0]] * k

            y0 = wrap(y)[0]
            _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
            x_std_fallback = nonecast(x_std, fill_value=1.0)
            return x, x_std_fallback, s


    def tsa_theta_auto(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_theta_factory(y=y,s=s,k=k,a=a,t=t,e=e, method='auto', deseasonalize=False)


    def tsa_theta_additive(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_theta_factory(y=y,s=s,k=k,a=a,t=t,e=e, method='additive', deseasonalize=False)


    def tsa_theta_multiplicative(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                               t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
            return tsa_theta_factory(y=y, s=s, k=k, a=a, t=t, e=e, method='multiplicative', deseasonalize=False)


    TSA_THETA_SKATERS = [ tsa_theta_auto, tsa_theta_additive, tsa_theta_multiplicative ]
else:
    TSA_THETA_SKATERS = []



if __name__=='__main__':
    assert using_statsmodels
    prior_plot_exogenous(f=tsa_theta_auto,k=1,n=TSA_META['n_warm']+25,n_plot=25)



from timemachines.skaters.pycrt.pycaretinclusion import using_pycaret
from timemachines.inclusion.pandasinclusion import using_pandas
from timemachines.skaters.sk.skinclusion import using_sktime
import datetime
from typing import List


if using_pycaret and using_pandas and using_sktime:
    import pandas as pd
    import numpy as np
    from pycaret.internal.pycaret_experiment import TimeSeriesExperiment
    from sktime.forecasting.base import ForecastingHorizon
    from timemachines.skatertools.utilities.suppression import no_stdout_stderr

    def pycrt_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, n_select=3, blend_method='median', turbo=True):
        """
            Uses pycaret on univariate series (for now)
        """
        # Examples: https://github.com/pycaret/pycaret/blob/time_series_beta/time_series_101.ipynb
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        # TODO: Modify this when t passed in
        y0_frame = pd.Series(index=pd.PeriodIndex(pd.date_range("2021-01", periods=len(y0s), freq="H")), data=y0s).to_frame()
        fold = 3
        with no_stdout_stderr():
            # Full fit cycle each time, for now
            exp = TimeSeriesExperiment()
            exp.setup(data=y0_frame, n_jobs=1, fh=k, verbose=False)
            best_baseline_models = exp.compare_models(fold=fold, sort='rmse', n_select=n_select,exclude=["auto_arima"], turbo=turbo)
            best_tuned_models = [exp.tune_model(model) for model in best_baseline_models]
            blended_model = exp.blend_models(best_tuned_models, method=blend_method)
            x = list(exp.predict_model(blended_model))
        x_std = [ 1 for _ in x ] # Will be overridden by empirical
        return x, x_std

if __name__=='__main__':
    assert using_pycaret, 'pip install pycaret-ts-alpha'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = pycrt_iskater(y=y, k=5)
    print('x')
    print(x)
    print('x_std')
    print(x_std)
    print('Elapsed')
    print(time.time()-st)


from timemachines.skaters.suc.successorinclusion import using_successor

if using_successor:
    from successor.skaters.scalarskaters.allscalarskaters import SCALAR_SKATERS
    SUCCESSOR_SKATERS = SCALAR_SKATERS
else:
    SUCCESSOR_SKATERS = []


from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
from timemachines.inclusion.statsmodelsinclusion import using_statsmodels

if using_statsmodels:

    try:
       from statsmodels.tsa.arima.model import ARIMA
    except ImportError:
       from statsmodels.tsa.arima_model import ARIMA

    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
    from typing import Any
    from timemachines.skaters.tsa.tsaparams import TSA_META
    from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
    from timemachines.skatertools.components.parade import parade
    from timemachines.skatertools.utilities.nonemath import nonecast
    import warnings
    from statsmodels.tools.sm_exceptions import ConvergenceWarning
    warnings.simplefilter('ignore', ConvergenceWarning)


    TSA_P_DEFAULT = 3
    TSA_D_DEFAULT = 0
    TSA_Q_DEFAULT = 3


    def tsa_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None,
                                 p:int=TSA_P_DEFAULT, d:int=TSA_D_DEFAULT, q:int=TSA_D_DEFAULT) -> ([float], Any, Any):
        """ Univariate, fixed p,d,q ARIMA model that is re-fit each time
            Reports empirical std
        """

        y = wrap(y)
        a = wrap(a)

        if not s.get('y'):
            s = {'y': list(),
                 'a': list(),
                 'k': k,
                 'p':{}}
        else:
            # Assert immutability of k, dimensions
            if s['y']:
                assert len(y) == len(s['y'][0])
                assert k == s['k']
            if s['a']:
                assert len(a) == len(s['a'][0])

        if y is None:
            return None, s, None
        else:
            s['y'].append(y)
            if a is not None:
                s['a'].append(a)
            if len(s['y']) > max(2 * k + 5, TSA_META['n_warm']):
                y0s = [ y_[0] for y_ in s['y']]
                model = ARIMA(y0s, order=(p,d,q))
                try:
                    x = list( model.fit().forecast(steps=k) )
                except:
                    x = [wrap(y)[0]]*k
            else:
                x = [y[0]] * k

            y0 = wrap(y)[0]
            _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
            x_std_fallback = nonecast(x_std, fill_value=1.0)
            return x, x_std_fallback, s




    def tsa_p1_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=0,q=0)


    def tsa_p2_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=0,q=0)


    def tsa_p3_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=0,q=0)

    TSA_D0_Q0_SKATERS = [ tsa_p1_d0_q0, tsa_p2_d0_q0,
                          tsa_p3_d0_q0 ]

    def tsa_p1_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=0,q=1)


    def tsa_p2_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=0,q=1)


    def tsa_p3_d0_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=0,q=1)

    TSA_D0_Q1_SKATERS = [ tsa_p1_d0_q1, tsa_p2_d0_q1,
                          tsa_p3_d0_q1 ]

    TSA_D0_SKATERS = TSA_D0_Q0_SKATERS + TSA_D0_Q1_SKATERS

    #################################### d=1 ####################################



    def tsa_p1_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=1,q=0)


    def tsa_p2_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=1,q=0)


    def tsa_p3_d1_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=1,q=0)


    TSA_D1_Q0_SKATERS = [ tsa_p1_d1_q0, tsa_p2_d1_q0,
                          tsa_p3_d1_q0 ]

    def tsa_p1_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=1,q=1)


    def tsa_p2_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=2,d=1,q=1)


    def tsa_p3_d1_q1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=3,d=1,q=1)


    TSA_D1_Q1_SKATERS = [ tsa_p1_d1_q0, tsa_p2_d1_q0,
                          tsa_p3_d1_q0 ]

    TSA_D1_SKATERS = TSA_D1_Q0_SKATERS + TSA_D1_Q1_SKATERS

    TSA_CONSTANT_SKATERS = TSA_D0_SKATERS + TSA_D1_SKATERS





    def tsa_balanced_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        """
             "less than precision" weighted moving averages
        """
        fs = TSA_CONSTANT_SKATERS
        return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.25)


    def tsa_precise_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        """
             Precision weight moving averages
        """
        fs = TSA_CONSTANT_SKATERS
        return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


    def tsa_aggressive_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        """
             Precision weight moving averages
        """
        fs = TSA_CONSTANT_SKATERS
        return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.75)


    def tsa_ma_precise_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        """
             Precision weight moving averages
        """
        from timemachines.skaters.simple.movingaverage import EMA_SKATERS
        fs = TSA_CONSTANT_SKATERS
        return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


    TSA_ENSEMBLE_SKATERS = [ tsa_aggressive_ensemble, tsa_balanced_ensemble, tsa_precise_ensemble ]

    TSA_SKATERS = TSA_ENSEMBLE_SKATERS + TSA_CONSTANT_SKATERS

else:
    TSA_D0_SKATERS = []
    TSA_D1_SKATERS = []
    TSA_CONSTANT_SKATERS = []
    TSA_ENSEMBLE_SKATERS = []
    TSA_SKATERS = []


if __name__=='__main__':
    assert using_statsmodels
    prior_plot_exogenous(f=tsa_p1_d0_q1,k=5,n=TSA_META['n_warm']+25,n_plot=50)
